#!/usr/bin/perl

# Convert a TERA reconciliation file to SylvX format.

use strict;
use teraUtils;

my $DEBUG = 0;

my $gAlphaName = "OUTGROUP";
my $gSpeciesRootName = "root";
my $gUndated = 0;
my $gEventRoot = 'root';
my $oldReconFormat = 0; # set to true if TTD/TFD use old ordering

my ($teraSpeciesFile, $reconFile, @opts) = @ARGV;
if( !$reconFile ) {
    print "USAGE: <newick_speciesFile> <reconFile> [-undated]$/";
    exit(1);
}

foreach my $opt (@opts) {
    if( $opt eq '-undated' ) {
        $gUndated = 1;
    } elsif( $opt eq '-d' ) {
        $DEBUG = 1;
    } elsif( $opt eq 'old' ) {
        $oldReconFormat = 1;
    } else {
        die "Unknown option: $opt";
    }
}

# event graph and intervals
my %gChildEvents;
my %gParentEvent;
my %gIntervals;

run($teraSpeciesFile, $reconFile);
print "SUCCESS$/";
exit(1);

sub run {
    my ($speciesFile, $reconFile ) = @_;

    # read the species trees and calculate dates for the internal nodes
    my ($speciesRootId,$speciesParentRef,$speciesChildrenRef,$branchLengthsRef) 
        = readTree( $teraSpeciesFile );
    $speciesParentRef->{$speciesRootId} = $gSpeciesRootName;

    my %speciesDates;
    getSpeciesDates( $speciesRootId, \%speciesDates, $speciesChildrenRef,
                     $branchLengthsRef );

    # output modified species tree
    printSpeciesTree( $teraSpeciesFile, \%speciesDates, $speciesParentRef );



    # read the TERA recon file
    my ($root,$eventListsRef,$geneParentsRef,$geneChildrenRef) 
        = teraUtils::readRecon( $reconFile );

    # make event graph 
    makeGraph( $root, $gEventRoot, $eventListsRef, $geneParentsRef, 
               $geneChildrenRef, $speciesParentRef, $speciesChildrenRef );

    # create initial time intervals for events
    my %restrictedTransfers; # transfers that have interval less than full tube
    makeIntervals( $gEventRoot, -1, \%restrictedTransfers, \%speciesDates,
                    $speciesParentRef );

    # traverse, giving max to children (e.g. transfers before losses)
    restrictInterval( $gEventRoot, \%restrictedTransfers );

    # set transfer points and refine event intervals
    my $transferPointsRef = restrictTubes( $gEventRoot, \%restrictedTransfers,
                    \%speciesDates );

    # sort genes and collect all events for each gene
    my %eventsByGene;
    my @geneOrder;
    my %seenIt;
    getEvents( $gEventRoot, \%seenIt, \@geneOrder, \%eventsByGene );


    # create sylvx file
    open NEW, ">$reconFile.mr" or die "Could not create SylvX mr file";
    print NEW "------------------------$/";
    print NEW "Rec $reconFile$/";

    printGraphEvents( \@geneOrder, \%eventsByGene, $transferPointsRef,
                      $geneParentsRef, $geneChildrenRef, $speciesChildrenRef );
    close NEW;
}



# return true if this event is the transfer to part of a transfer
sub isTransferTo {
    my ($ge) = @_;

    my ($geneId,$tube,$type) = 
            $ge =~ /^(.+):(\(.+,.+\)):(.+)$/;

    die "isTransferTo called on non-T event" if $type ne 'T';

    my $parentEvent = $gParentEvent{$ge};
    my ($p,$parentTube,$pType) = 
            $parentEvent =~ /^(.+):(\(.+,.+\)):(.+)$/;
    if( !$parentTube || $parentTube ne $tube ) {
        return 1;
    } else {
        return 0;
    }
}

# sort by first part of interval
sub sortByStart {
    return $gIntervals{$b}[0] <=> $gIntervals{$a}[0];
}

# finish - print events to NEW file handle
sub printGraphEvents {
    my ($geneOrderRef, $eventsByGeneRef, $transferPointsRef,
        $geneParentsRef,$geneChildrenRef,$speciesChildrenRef) = @_;

    my %eventEnds; # for linking events
    foreach my $geneId (@$geneOrderRef) {
        my $geneParentId = $geneParentsRef->{$geneId};
        $geneParentId = 0 if( !$geneParentId );

        # print gene pair
        print "{$geneParentId,$geneId}$/" if $DEBUG;
        print NEW "{$geneParentId,$geneId}$/";
        my $firstPass = 1;
        foreach my $ge ( sort sortByStart @{$eventsByGeneRef->{$geneId}}) {
            my ($geneId,$tube,$type) = $ge =~ /^(.+):(\(.+,.+\)):(.+)$/;
            print "   $ge $gIntervals{$ge}[0]-$gIntervals{$ge}[1]$/"
                if $DEBUG;
            my $len = $gIntervals{$ge}[0]-$gIntervals{$ge}[1];
            my $mid = $gIntervals{$ge}[1] + $len/2;
            my $quarter = $gIntervals{$ge}[0] - $len/4;
            my $parentEvent = $gParentEvent{$ge};
            my ($p,$parentTube,$pType) = 
                    $parentEvent=~ /^(.+):(\(.+,.+\)):(.+)$/;

            # prevents hanging losses
            if( $firstPass 
                && @{$eventsByGeneRef->{$geneId}} > 1
                && ($geneId =~ /.+-$/ || $parentEvent eq 'root' ) ) {
                print NEW "\t$tube [$gIntervals{$ge}[0]-$gIntervals{$ge}[0]]$/";
            }

            # match up to previous
            my $start = $gIntervals{$ge}[0];
            if( $type ne 'S' && $parentTube eq $tube && 
                $start < $eventEnds{$parentEvent} ) 
            {
                $start = $eventEnds{$parentEvent};
            }
            my $end = $gIntervals{$ge}[1];

            # print interval
            my $interval = "[$start-$end]";
            if( $type eq 'S' || $type eq 'D' || $type eq 'I' ) {
                print NEW "\t$tube $interval$/"; # just full interval 
            } elsif( $type eq 'L' ) {
                # quarter interval
                $interval = "[$start-$end:$start-$quarter]";
                print NEW "\t$tube $interval$/";
            } elsif( $type eq 'T' ) {

                # set transfer point to mid interval if not already set
                my $transferPoint = $mid;
                if( $transferPointsRef->{$ge} ) {
                    $transferPoint = $transferPointsRef->{$ge};
                }

                # start with from species if it is the transfered to event
                if( $parentTube && $parentTube ne $tube ) {
                    #isTransferTo
                    $interval = "[$transferPoint-$end:$transferPoint]";
                    print NEW "\t$parentTube $interval$/";
                }

                if( $tube eq "($gAlphaName,$gAlphaName)" ) {
                    # no lines in outgroup, otherwise line is bi-dir 
                    my $tpInc = $transferPoint-0.01;
                    $interval .= " [$transferPoint-$tpInc]"; 
                } elsif( !$geneChildrenRef->{$geneId} ) {
                    my ($parentSpecies, $species) = $tube =~ /\((.*),(.*)\)/;
                    die "Failed to parse tube" if( !$species );
                    if( !$speciesChildrenRef->{$species} ) { 
                        $interval = " [$transferPoint-0]"; # brings line to end
                    }
                } elsif( $parentTube ne $tube ) {
                    # transfer to
                    my $tpInc = $transferPoint-0.01;
                    if( $gChildEvents{$ge} ) {
                        # connect with child
                        my $child = $gChildEvents{$ge}[0];
                        $tpInc = $gIntervals{$child}[0];
                    }
                    $interval .= " [$transferPoint-$tpInc]"; 
                }
                print NEW "\t$tube $interval$/";
            } else {
                die "Bad type for event: $ge";
            }
            $eventEnds{$ge} = $end;
            $firstPass = 0;
        }
    }
}

# get a recursive gene order and collect events for each gene
sub getEvents {
    my ($graphEvent,$seenItRef,$geneOrderRef,$eventsByGeneRef) = @_;
    my ($geneId,$spParent,$spChild,$type) = 
            $graphEvent =~ /^(.+):\((.+),(.+)\):(.+)$/;

    if( $graphEvent ne $gEventRoot && !$seenItRef->{$geneId} ) {
        push @$geneOrderRef, $geneId;
        $seenItRef->{$geneId} = 1;
    }

    push @{$eventsByGeneRef->{$geneId}}, $graphEvent;
    foreach my $child (@{$gChildEvents{$graphEvent}}) {
        getEvents( $child, $seenItRef, $geneOrderRef, $eventsByGeneRef );
    }
}


# - get a tube order (recursive)
# - get all events in tube
# - start new tube if parent event is in different tube
# - only first duplication event
sub makeTubes {
    my ($graphEvent, $tubeOrderRef, $tubesRef, $parentTube, $idx, $child ) = @_;

    my ($geneId,$tube,$type) = 
            $graphEvent =~ /^(.+):(\(.+,.+\)):(.+)$/;

    if( $graphEvent ne $gEventRoot 
        && $geneId !~ /.+-$/ )  # ignore loss ids and speciations
    {
        if( $tube ne $parentTube ) {
            #start a new tube by finding non-used name
            $idx = 0;
            while( $tubesRef->{"$tube-$idx"} ) {
                $idx++;
            }
        }
        if( $type ne 'D' || $child == 0 ) {
            # only first duplication
            my $tubeName = "$tube-$idx";
            push @$tubeOrderRef, $tubeName if( !$tubesRef->{$tubeName} );
            push @{$tubesRef->{$tubeName}}, $graphEvent;
        }
    }

    my $i=0;
    foreach my $e (@{$gChildEvents{$graphEvent}}) {
        makeTubes( $e, $tubeOrderRef, $tubesRef, $tube, $idx, $i );
        $i+=1;
    }
}

# get sibling event
sub getSibling {
    my ($graphEvent) = @_;

    if( !$gParentEvent{$graphEvent} ) {
        die "Trying to get sibling of root";
    }

    my $pe = $gParentEvent{$graphEvent};
    if( @{$gChildEvents{$pe}} != 2 ) {
        return ''; # no sibling
    }

    if( $graphEvent eq $gChildEvents{$pe}[0] ) {
        return $gChildEvents{$pe}[1];
    } elsif( $graphEvent eq $gChildEvents{$pe}[1] ) {
        return $gChildEvents{$pe}[0];
    } else {
        die "$graphEvent not found in children of $pe$/";
    }
}
    

# set a transfer point for the event and it's sibling
# then propagate it as a maximum to the children events
sub setTransferPoint {
    my ($graphEvent, $point, $transferPointsRef) = @_;
   
    my $sib = getSibling( $graphEvent );
    if( $sib && $transferPointsRef->{$sib} 
        && $transferPointsRef->{$sib} != $point ) 
    {
        die "Calling set transfer point ($point)"
             . " for $graphEvent when sibling already"
              . " set $sib at $transferPointsRef->{$sib}";
    }

    $transferPointsRef->{$graphEvent} = $point;
    $gIntervals{$graphEvent}[0] = $point;
    if( $sib ) {
        $transferPointsRef->{$sib} = $point;
        $gIntervals{$sib}[0] = $point;
    }
    print "  SET TRANSFER $graphEvent $point (and for sib $sib)$/" if $DEBUG;

    # propagate sibling (transferred to)
    my @q;
    push @q, @{$gChildEvents{$sib}} if $sib;
    push @q, @{$gChildEvents{$graphEvent}}; 
    while( @q ) {
        my $ge = shift @q;
        if( $gIntervals{$ge}[0] > $point ) {
            print "  PROPOGATE $graphEvent $point to $ge$/" if $DEBUG;
            if( $point < $gIntervals{$ge}[1] ) {
                die "ERROR $ge: propagating $point past interval min"
                      . " ($gIntervals{$ge}[1])$/";
            }
            $gIntervals{$ge}[0] = $point;
            push @q, @{$gChildEvents{$ge}};
        }
    }
}

# return true if the event is the last in it's tube
sub isLast {
    my ($ge) = @_;
    
    my ($geneId,$tube,$type) = $ge =~ /^(.+):(\(.+,.+\)):(.+)$/;

    my $lastChild = 1;
    foreach my $e (@{$gChildEvents{$ge}}) {
        my ($g,$childTube,$t) = $e =~ /^(.+):(\(.+,.+\)):(.+)$/;
        if( $g !~ /.+-$/   # ignore loss ids and speciations
            && $tube eq $childTube ) {
            $lastChild = 0;
        }
    }

    return $lastChild;
}


# extend the lower interval of the last children in a tube to complete
# tube
sub extendLastChildren {
    my ($speciesDatesRef,@curTube) = @_;

    return if @curTube < 1;

    my ($geneId,$tube,$type) = $curTube[0] =~ /^(.+):(\(.+,.+\)):(.+)$/;
    my ($spParent,$spChild) = $tube =~ /\((.+),(.+)\)/;
    my $tubeEnd = $speciesDatesRef->{$spChild}[0];

    foreach my $ge (@curTube) {
        my ($geneId,$tube,$type) = $curTube[0] =~ /^(.+):(\(.+,.+\)):(.+)$/;
        if( isLast( $ge ) ) {
            $gIntervals{$ge}[1] = $tubeEnd;
            print "EXTENDING $ge in $tube to $tubeEnd$/" if $DEBUG;
        }
        if( $type eq 'D' ) {
            # check sibling
            my $sib = getSibling( $ge );
            die "extendLastChildren: no sibling" if( !$sib );
            if( isLast( $sib ) ) {
                $gIntervals{$sib}[1] = $tubeEnd;
                print "EXTENDING sib $sib in $tube to $tubeEnd$/" if $DEBUG;
            }
        }
    }
}

# chain the intervals in a tube
sub tubeProc {
    my ($transferPointsRef,$speciesDatesRef,$tube, $start, $end, @curTube) 
            = @_;

    die "no tubes for tubeProc" if !@curTube;

    if( @curTube < 2 ) {
        my $ge = $curTube[0];
        print " >>$ge $gIntervals{$ge}[0]-$gIntervals{$ge}[1] NO CHANGE$/"
            if $DEBUG;
        return;
    }


    my ($spParent,$spChild) = $tube =~ /\((.+),(.+)\)/;
    my $tubeEnd = $speciesDatesRef->{$spChild}[0];

    # set start and end to max and min of all events
    $start = $gIntervals{$curTube[0]}[0];
    $end = $gIntervals{$curTube[0]}[1];
    foreach my $ge (@curTube) {
        if( $start < $gIntervals{$ge}[0] ) {
            $start = $gIntervals{$ge}[0];
        }
        if( $end > $gIntervals{$ge}[1] ) {
            $end = $gIntervals{$ge}[1];
        }
    }

    my $inc = ($end-$start)/@curTube;
    print "TUBE PROC $tube: [$start-$end] inc=$inc " . @curTube 
        . " elements$/" if $DEBUG;

    # chain events
    my $nextStart = $start;
    for( my $i=0; $i<@curTube; $i++ ) {
        my $ge = $curTube[$i];
        my ($geneId,$tube,$type) = $ge =~ /^(.+):(\(.+,.+\)):(.+)$/;

        $gIntervals{$ge}[0] = $nextStart if( $i>0 ); # keep start
        $nextStart += $inc;
        $nextStart = 0 if $nextStart < 0.001;
        $gIntervals{$ge}[1] = $nextStart;

        if( ($type eq 'T' && !$transferPointsRef->{$ge} )
            || $type eq 'D' ) 
        {  # match siblings
            my $sib = getSibling( $ge );
            if( $sib ) {
                $gIntervals{$sib}[0] = $gIntervals{$ge}[0];
                $gIntervals{$sib}[1] = $gIntervals{$ge}[1];
                print " >sib $sib $gIntervals{$ge}[0]-$gIntervals{$ge}[1]$/"
                    if $DEBUG;
                if( $type eq 'T' ) {
                    # set transfer point to beginning if not already set
                    setTransferPoint( $ge, $gIntervals{$ge}[0], 
                                        $transferPointsRef);
                }
            } elsif( $type eq 'D' ) {
                die "tubeProc: no sibling for duplicate";
            }
        }
            
        print " >$ge $gIntervals{$ge}[0]-$gIntervals{$ge}[1]$/" if $DEBUG;
    }
}


# set transfer points and intervals for each event
sub restrictTubes {

    my ($root, $restrictedTransfersRef,$speciesDatesRef) = @_;

    # get all events in each species tube
    my @tubeOrder;
    my %tubes;
    makeTubes( $root, \@tubeOrder, \%tubes );

    my %transferPoints;

    # process each tube
    foreach my $tube (@tubeOrder) {

        my @curTube = @{$tubes{$tube}};
        next if( @curTube == 0 );

        my ($spParent,$spChild) = $tube =~ /\((.+),(.+)\)/;
        my $start = $speciesDatesRef->{$spParent}[0];
        my $end = $speciesDatesRef->{$spChild}[0];

        if( $DEBUG ) {
            print "====TUBE $tube " . @curTube . " elements [$start-$end]$/";
            foreach my $ge (@curTube) {
                print "      $ge [$gIntervals{$ge}[0]-$gIntervals{$ge}[1]]$/"
            }
        }

        next if( @curTube == 1 );

        # Loop through events in tube.
        # Set transfer points.
        # Order non-transfer events between the transfer points 
        #  (or tube start/end).
        my @nonTransferEvents; 
        my $prevPoint = $start;
        my $previousTransfer;
        my $transPoint;
        for( my $i=0; $i<=@curTube; $i++ ) {
            my $ge;
            if( $i<@curTube ) {
                $ge = $curTube[$i];
                $transPoint = $transferPoints{$ge};
            }

            # get/set transfer point
            if( !$ge ) {
                # last pass, no event
                $transPoint = $end;
            } elsif( $transferPoints{$ge} ) {
                #print "T: SET $ge $gIntervals{$ge}[0]"
                #      . "-$gIntervals{$ge}[1] tp=$transferPoints{$ge}$/";
            } elsif( $restrictedTransfersRef->{$ge} || $transferPoints{$ge} ) {
                print "T:RESTRICTED $ge $gIntervals{$ge}[0]"
                      . "-$gIntervals{$ge}[1]$/" if $DEBUG;
                my $len = $gIntervals{$ge}[0] -$gIntervals{$ge}[1];
                $transPoint = $gIntervals{$ge}[1] + $len/2;
                setTransferPoint( $ge, $transPoint, \%transferPoints );
            } else {
                # not a transfer, continue
                print "   $ge$/" if $DEBUG;
                push @nonTransferEvents, $ge;
                next;
            }

            if( $previousTransfer && !isTransferTo( $previousTransfer ) ) {
                $gIntervals{$previousTransfer}[1] = $transPoint;
            }

            if( @nonTransferEvents) {
                # process non-transfer events
                tubeProc( \%transferPoints, $speciesDatesRef, $tube, 
                                $prevPoint, $transPoint, @nonTransferEvents );
                @nonTransferEvents = ();
            }
            $prevPoint = $transPoint;
            $previousTransfer= $ge;
        }
       
        # Possible hack. This should work in tubeProc.
        # make sure tube is complete
        extendLastChildren( $speciesDatesRef, @curTube ); 
    }

    return \%transferPoints;
}

# propogate minimums up and sync intervals for transfers and duplicates
sub restrictInterval {
    my ($graphEvent,$restrictedTransfersRef) = @_;

    my ($geneId,$spParent,$spChild,$type) = 
            $graphEvent =~ /^(.+):\((.+),(.+)\):(.+)$/;

    foreach my $e (@{$gChildEvents{$graphEvent}}) {
        restrictInterval( $e, $restrictedTransfersRef );
    }

    if( $graphEvent ne $gEventRoot ) {
        foreach my $e (@{$gChildEvents{$graphEvent}}) {
            # give children min to parents if greater 
            if( $gIntervals{$graphEvent}[1] < $gIntervals{$e}[1] )
            {
                print "min $graphEvent $gIntervals{$graphEvent}[1]"
                        . " -> $gIntervals{$e}[1] from child=$e$/" if $DEBUG;
                $gIntervals{$graphEvent}[1] = $gIntervals{$e}[1];
                if( $gIntervals{$graphEvent}[1] > $gIntervals{$graphEvent}[0] 
                    && $spParent ne $gAlphaName )
                {
                    die "restrictMinInterval too much for $graphEvent";
                }
                if( $type eq 'T' || $type eq 'D' ) {
                    # propagate to sibling
                    my $sib = getSibling( $graphEvent );
                    if( $sib ) { 
                        $restrictedTransfersRef->{$graphEvent} = $sib 
                            if $type eq 'T';
                        if( $gIntervals{$sib}[1] < $gIntervals{$e}[1] ) {
                            print "min sib $sib $gIntervals{$sib}[1] ->"
                                . " $gIntervals{$e}[1] from child=$e$/" 
                                if $DEBUG;
                            $gIntervals{$sib}[1] = $gIntervals{$e}[1];
                            $restrictedTransfersRef->{$sib} = $graphEvent 
                                if $type eq 'T';
                        }
                        my ($sibGeneId,$sibSpParent,$sibSpChild,$sibType) = 
                                $sib =~ /^(.+):\((.+),(.+)\):(.+)$/;
                        if( $gIntervals{$sib}[1] > $gIntervals{$sib}[0] 
                            && $sibSpParent ne $gAlphaName ) 
                        {
                            die "restrictMinInterval too much for sib $sib";
                        }
                    } elsif( $type eq 'D' ) {
                        die "restrictInterval: no sibling for duplicate";
                    }
                }
            }
        }
    }
}



# ISDL events are species tube length
# T events are intersection of siblings 
# events = <geneId>:(spParent,spChild):[DTLS]
sub makeIntervals {
    my ($graphEvent,$parentMax,$restrictedTransfersRef,$speciesDatesRef,
        $speciesParentRef) = @_;

    my ($geneId,$spParent,$spChild,$type) = 
        $graphEvent =~ /^(.+):\((.+),(.+)\):(.+)$/;

    # set initial intervals
    if( $graphEvent ne $gEventRoot 
        && !$gIntervals{$graphEvent} ) {  # T events fill siblings
        
        die "Bad event in makeIntervals $graphEvent" if( !$geneId );
  
        if( $type eq 'T' ) {
            # get sibling event
            my $siblingEvent;
            foreach my $e (@{$gChildEvents{$gParentEvent{$graphEvent}}}) {
                $siblingEvent = $e if( $e ne $graphEvent );
            }

            my $max = $speciesDatesRef->{$speciesParentRef->{$spChild}}[0];
            if( $parentMax != -1 && $parentMax < $max ) {
                print "MAX from parent = $parentMax for $graphEvent$/"
                    if $DEBUG;
                $max = $parentMax;
                $restrictedTransfersRef->{$graphEvent} = $siblingEvent;
            }
            my $min = $speciesDatesRef->{$spChild}[0];
            if( $max<$min ) {
                print "interval: "
                . "$speciesDatesRef->{$speciesParentRef->{$spChild}}[0]-$min$/";
                die "Impossible parent ordering for $graphEvent$/";
            }

            if( $siblingEvent ) { # outgroups don't appear

                my ($sibGeneId,$sibOldParent,$spSibling,$sibType) = 
                    $siblingEvent =~ /^(.*):\((.*),(.*)\):(.*)$/;
                my $sibMax = $speciesDatesRef->
                    {$speciesParentRef->{$spSibling}}[0];
                my $sibMin = $speciesDatesRef->{$spSibling}[0];

                # create intersection
                my $origMax = $max;
                my $origMin = $min;
                if( $sibMax < $max ) {
                    $max = $sibMax;
                    $restrictedTransfersRef->{$graphEvent} = $siblingEvent;
                } elsif( $max < $sibMax ) {
                    $restrictedTransfersRef->{$siblingEvent} = $graphEvent;
                }
                if( $sibMin > $min ) {
                    $min = $sibMin; 
                    $restrictedTransfersRef->{$graphEvent} = $siblingEvent;
                } elsif( $min > $sibMin ) {
                    $restrictedTransfersRef->{$siblingEvent} = $graphEvent;
                }

                if( $max<$min ) {
                    if( !$gUndated ) {
                        print "interval: $siblingEvent $max<$min$/";
                        die "Impossible ordering for $graphEvent$/";
                    }
                    # undated trees don't necessarily intersect use greater
                    if( $origMax > $sibMax ) {
                        $max = $origMax;
                        $min = $origMin;
                    } else {
                        $max = $sibMax;
                        $min = $sibMin;
                    }
                } 

                # set sibling
                $gIntervals{$siblingEvent}[0] = $max;
                $gIntervals{$siblingEvent}[1] = $min;
            }

            $gIntervals{$graphEvent}[0] = $max;
            $gIntervals{$graphEvent}[1] = $min;

        } else {
            # other events - in
            $gIntervals{$graphEvent}[0] = $speciesDatesRef->{$spParent}[0];
            $gIntervals{$graphEvent}[1] = $speciesDatesRef->{$spChild}[0];
        }
    }

    print "$parentMax interval $graphEvent = "
        . "$gIntervals{$graphEvent}[0] - $gIntervals{$graphEvent}[1]$/"
        if $DEBUG;
    foreach my $e (@{$gChildEvents{$graphEvent}}) {
        print "  $graphEvent --> $e$/" if $DEBUG;
        my $max = $gIntervals{$graphEvent}[0]+0.001;
        if( $graphEvent eq $gEventRoot || $spParent eq $gAlphaName ) {
            $max = -1;
        }
        makeIntervals( $e, $max, $restrictedTransfersRef, $speciesDatesRef,
                        $speciesParentRef );
    }
}




sub getNextSpecies {
    my( $i, $event, $firstType, $secondType, $geneId, $spParent, $lostChild,
        $eventListsRef, $speciesChildrenRef) = @_;
    my $nextSpecies;
    if( $secondType eq 'LTD' ) {
        $nextSpecies = $gAlphaName; # next event is to dead
    } elsif( $i+1 == @{$eventListsRef->{$geneId}} ) {
        # last event
# This assumes gene name separator is '_'
        if( $firstType eq 'T' ) {
            if( $geneId =~ /^([^_]+)_(.*)/ ) {
                $nextSpecies = $1;
            } else {
                die "Cannot get TLFD target for $geneId : $event";
            }
        } elsif( $speciesChildrenRef->{$spParent}[0] eq $lostChild ) {
            $nextSpecies = $speciesChildrenRef->{$spParent}[1];
        } elsif( $speciesChildrenRef->{$spParent}[1] eq $lostChild ) {
            $nextSpecies = $speciesChildrenRef->{$spParent}[0];
        } else {
            print "species parent=$spParent$/";
            foreach my $c (@{$speciesChildrenRef->{$spParent}}) {
                print "Child: $c$/";
            }
            die "Could not find lost child ($lostChild) for "
                . "$geneId:$event";
        }
    } else {
        # get it from next event
        my $nextEvent = $eventListsRef->{$geneId}[$i+1];
        if( $nextEvent =~ /^([^,]+),.*/ ) {
            $nextSpecies = $1;
        } else {
            die "Could not read next event: $nextEvent";    
        }
    }
    return $nextSpecies;
}



# TERA doesn't order alpha events.
# Return true if the first gene's event is from alpha (-1)
sub alphaFirst {
    my ($gene1, $gene2,$eventListsRef) = @_;

    if( $eventListsRef->{$gene1} ) {
        return 1 if( $eventListsRef->{$gene1}[0] =~ /^-1,.*/ )
    } 
    if( $eventListsRef->{$gene2} ) {
        return 0 if( $eventListsRef->{$gene2}[0] =~ /^-1,.*/ )
    } 

    die "alphaFirst did not find alpha event";
}

# Make an event graph from the TERA recon.
# fill tubes, original internal, parent event (constraint)
sub makeGraph {
    my ($geneId,$parentEvent,$eventListsRef,$geneParentsRef,
        $geneChildrenRef,$speciesParentRef,$speciesChildrenRef) = @_;
    return if !$eventListsRef->{$geneId};

    for( my $i=0; $i<@{$eventListsRef->{$geneId}}; $i++ ) {
        my $event = $eventListsRef->{$geneId}[$i];
        if( $event =~ /^(.*),([SDT]),(.*),(.*)@/ ) {
            my ($spParent,$type,$spChild1,$spChild2) = ($1,$2,$3,$4);
            if( !$geneChildrenRef->{$geneId} ) {
                die "SDT $geneId:$event doesn't have children";
            }
            if( $i != @{$eventListsRef->{$geneId}}-1 ) {
                die "SDT $geneId:$event not last";
            }

            my $spParent1 = $speciesParentRef->{$spChild1};
            my $spParent2 = $speciesParentRef->{$spChild2};
            if( $spParent1 eq $gSpeciesRootName && $spChild2 eq 'alpha' 
                && $type eq 'S' ) 
            {
                $spChild2 = $gAlphaName;
                $spParent2 = $gSpeciesRootName;
            } elsif( $spParent2 eq $gSpeciesRootName && $spChild1 eq 'alpha' 
                && $type eq 'S' )  
            {
                $spChild1 = $gAlphaName;
                $spParent1 = $gSpeciesRootName;
            } elsif( !$spParent1 ) {
                die "makeGraph to species parent for $spChild1";
            } elsif( !$spParent2 ) {
                die "makeGraph to species parent for $spChild2";
            }   

            my $childEvent1 = "$geneChildrenRef->{$geneId}[0]:"
                         . "($spParent1,$spChild1):$type";
            my $childEvent2 = "$geneChildrenRef->{$geneId}[1]:"
                         . "($spParent2,$spChild2):$type";

            # add to graph
            $gParentEvent{$childEvent1} = $parentEvent;
            $gParentEvent{$childEvent2} = $parentEvent;
            push @{$gChildEvents{$parentEvent}}, $childEvent1;
            push @{$gChildEvents{$parentEvent}}, $childEvent2;
            makeGraph( $geneChildrenRef->{$geneId}[0], $childEvent1, 
                       $eventListsRef, $geneParentsRef, $geneChildrenRef,
                        $speciesParentRef, $speciesChildrenRef );
            makeGraph( $geneChildrenRef->{$geneId}[1], $childEvent2, 
                       $eventListsRef, $geneParentsRef, $geneChildrenRef,
                        $speciesParentRef, $speciesChildrenRef );
        } elsif( $event =~ /^(.*),I,(.*),(.*)@/ ) {
            my ($spParent,$spChild1,$spChild2) = ($1,$2,$3,$4);
            if( !$geneChildrenRef->{$geneId} ) {
                die "I $geneId:$event doesn't have children";
            }
            if( $i != @{$eventListsRef->{$geneId}}-1 ) {
                die "I $geneId:$event not last";
            }

            my $idx = 0;
            foreach my $spChild ($spChild1, $spChild2) {
                my $geneChild = $geneChildrenRef->{$geneId}[$idx];

                # first event is ILS
                my $parentParent = $speciesParentRef->{$spParent};
                my $childEvent = "$geneChild:($parentParent,$spParent):I";

                #add to graph
                $gParentEvent{$childEvent} = $parentEvent;
                push @{$gChildEvents{$parentEvent}}, $childEvent;

                # make bottom of list of descendants
                my @descendants;
                my $sp = $spChild;
                while( $sp ne $spParent ) { 
                    unshift @descendants, $sp;
                    if( $speciesParentRef->{$sp} ) {
                        $sp = $speciesParentRef->{$sp};
                    } else {
                        die "No matching parent from $sp";
                    }
                }

                # process parent list
                while( @descendants) {
                    my $child = shift @descendants;
                    $parentEvent = $childEvent;

                    my $parent = $speciesParentRef->{$child};
                    $childEvent = "$geneChild:($parent,$child):I";

                    #add to graph
                    $gParentEvent{$childEvent} = $parentEvent;
                    push @{$gChildEvents{$parentEvent}}, $childEvent;
                }

              
                makeGraph( $geneChild, $childEvent, 
                       $eventListsRef, $geneParentsRef, $geneChildrenRef,
                        $speciesParentRef, $speciesChildrenRef );

                $idx++;
            }
        } elsif ( $event =~ /^(.*),(T)TD,(.*),(.*)@/ 
                || $event =~ /^(.*),(T)FD,(.*),(.*)@/ 
                || $event =~ /^(.*),(D)D,(-1),(-1)@/ ) 
        {
            # alpha (outgroup) events
            my ($spParent,$type,$spChild1,$spChild2) = ($1,$2,$3,$4);
            if( !$geneChildrenRef->{$geneId} ) {
                die "TT/FD $geneId:$event doesn't have children";
            }
            if( $i != @{$eventListsRef->{$geneId}}-1 ) {
                die "TT/FD $geneId:$event not last";
            }

            # make sure children are in the right order
            my $geneChild1 = $geneChildrenRef->{$geneId}[0];
            my $geneChild2 = $geneChildrenRef->{$geneId}[1];
            if( $oldReconFormat &&  
                alphaFirst( $geneChild1, $geneChild2, $eventListsRef ) ) {
                # species children ordered opposite of gene children
                my $tmp = $geneChild1;
                $geneChild1 = $geneChild2;
                $geneChild2 = $tmp;
            }

            # change -1 to alpha name
            my $spParent1;
            if( $spChild1 eq '-1' ) {
                $spChild1 = $gAlphaName;
                $spParent1 = $gAlphaName;
            } else {
                $spParent1 = $speciesParentRef->{$spChild1};
            }

            my $spParent2;
            if( $spChild2 eq '-1' ) {
                $spChild2 = $gAlphaName;
                $spParent2 = $gAlphaName;
            } else {
                $spParent2 = $speciesParentRef->{$spChild2};
            }

            if( !$spParent2 ) {
                die "No second parent (child=$spChild2) for $geneId : $event$/";
            }

            my $childEvent1 = "$geneChild1:"
                         . "($spParent1,$spChild1):$type";
            my $childEvent2 = "$geneChild2:"
                         . "($spParent2,$spChild2):$type";

            # add to graph
            $gParentEvent{$childEvent1} = $parentEvent;
            $gParentEvent{$childEvent2} = $parentEvent;
            push @{$gChildEvents{$parentEvent}}, $childEvent1;
            push @{$gChildEvents{$parentEvent}}, $childEvent2;
            makeGraph( $geneChild1, $childEvent1, $eventListsRef,
                       $geneParentsRef, $geneChildrenRef, $speciesParentRef,
                         $speciesChildrenRef );
            makeGraph( $geneChild2, $childEvent2, $eventListsRef,
                       $geneParentsRef, $geneChildrenRef, $speciesParentRef,
                         $speciesChildrenRef );

        } elsif ( $event =~ /^(.*),(S)(L),(.*)@/ 
                ||  $event =~ /^(.*),(T)(L),(.*)@/  
                ||  $event =~ /^(-1),(T)(LFD),(.*)@/  
                ||  $event =~ /^(.*),(T)(LTD),(.*)@/ ) 
        {
            # loss events
            my ($spParent,$firstType,$secondType,$children) = ($1,$2,$3,$4);

            my $nextSpecies;
            my $lostChild;
            if( $oldReconFormat ) {
                $lostChild = $children;
            } else {
                ($lostChild,$nextSpecies) = $children =~ /^(.*),(.*)$/;
                if( !$nextSpecies ) {
                    die "Missing children: $geneId:$event$/";
                }
            }

            $lostChild = $gAlphaName if( $lostChild eq '-1' );
            $spParent = $gAlphaName if( $spParent eq '-1' );

            # get next non-loss species species id
            if( $oldReconFormat ) {
                $nextSpecies = getNextSpecies( $i, $event,
                    $firstType, $secondType, $geneId, $spParent, 
                    $lostChild, $eventListsRef, $speciesChildrenRef );
            }
            $nextSpecies = $gAlphaName if( $nextSpecies eq '-1' );

            if( !$geneParentsRef->{$geneId} ) {
                # skip SL to dead at root
                next if( $firstType eq 'S' 
                    && ( $lostChild eq $gAlphaName || $lostChild eq 'alpha') );
            }

            # add kept child
            my $keptParent = $speciesParentRef->{$nextSpecies};
            my $childEvent = "$geneId:($keptParent,$nextSpecies):$firstType";
            $gParentEvent{$childEvent} = $parentEvent;
            push @{$gChildEvents{$parentEvent}}, $childEvent;

            # add lost child
            if( $spParent ne $gAlphaName ) {
                my $lostGeneId = "$geneId-";
                $geneParentsRef->{$lostGeneId} = $geneParentsRef->{$geneId};
                
                my $lostChildEvent;
                if( $firstType eq 'T' ) {
                    $lostChildEvent 
                        = "$lostGeneId:($spParent,$lostChild):$firstType";
                } elsif( $speciesParentRef->{$spParent} ne $gSpeciesRootName ) {
                    $lostChildEvent = "$lostGeneId:"
                        . "($speciesParentRef->{$spParent},$spParent)"
                        . ":$firstType";
                }
                if( $lostChildEvent ) {
                    $gParentEvent{$lostChildEvent} = $parentEvent;
                    push @{$gChildEvents{$parentEvent}}, $lostChildEvent;
                }

                # lost child
                my $lossEvent = "$lostGeneId:($spParent,$lostChild):L";
                $gParentEvent{$lossEvent} = $parentEvent;
                push @{$gChildEvents{$parentEvent}}, $lossEvent;
            }

            $parentEvent = $childEvent;
        } else {
            die "Unknown event $geneId:$event$/";
        }
    }
}



# calculate the date intervals for each species
# [0] = minimum (leaves)
# [1] = maximum (root)
sub getSpeciesDates {
    my ($id,$speciesDatesRef,$speciesChildrenRef,$branchLengthsRef) = @_;

    if( !$speciesChildrenRef->{$id} ) {
        #leaf
        $speciesDatesRef->{$id}[0] = 0;
        $speciesDatesRef->{$id}[1] = $branchLengthsRef->{$id};
        return $branchLengthsRef->{$id};
    }

    # internal node
    my $childId0 = $speciesChildrenRef->{$id}[0];
    my $childId1 = $speciesChildrenRef->{$id}[1];
    # should be same length
    my $childLen = getSpeciesDates( $childId0, $speciesDatesRef, 
                        $speciesChildrenRef, $branchLengthsRef );
    my $len1 = getSpeciesDates( $childId1, $speciesDatesRef,
                        $speciesChildrenRef, $branchLengthsRef );
    if( abs($len1-$childLen) > 0.01 ) {
        die "Lengths different: $id=$childLen, $len1 "
            . " children = $childId0 and $childId1$/";
    }
    
    $speciesDatesRef->{$id}[0] = $childLen;
    my $len = $childLen + $branchLengthsRef->{$id};
    $speciesDatesRef->{$id}[1] = $len;

    return $len;
}


# print the modified species tree (with outgroup)
sub printSpeciesTree {
    my ($treeFile,$speciesDatesRef,$speciesParentRef) = @_;

    my $treeString;
    my $rootId;
    open TREE, $treeFile or die "Could not open $treeFile";
    while( <TREE> ) {
        chomp;
        next if /^\s*$/; #ignore blank lines
        if( /^\((.*)\)(\d+);\s*$/ ) {
            $treeString = $1;
            $rootId = $2;
            last;
        } elsif( /^(.*);\s*$/ ) {
            die "Bad format for species file - missing root id or parentheses.";
        } else {
            die "Species tree string not found";
        }
    }
    close TREE;

    # print species tree with outgroup
    open NEW_FILE, ">$treeFile" . "_OUT" 
        or die "Could not create outgroup file";
    my $alphaHeight = $speciesDatesRef->{$rootId}[0]+1;
    $speciesDatesRef->{$gAlphaName}[0] = 0;
    $speciesDatesRef->{$gAlphaName}[1] = $alphaHeight;
    $speciesDatesRef->{$rootId}[1] = $alphaHeight;
    $speciesDatesRef->{$gSpeciesRootName}[0] = $alphaHeight;
    $speciesParentRef->{$gAlphaName} = $rootId;
    print NEW_FILE "(($treeString)$rootId:1,$gAlphaName:$alphaHeight)"
        . "$gSpeciesRootName;$/";
    close NEW_FILE;
}

# read the species tree
sub readTree {
    my ($treeFile) = @_;

    my $treeString;
    open TREE, $treeFile or die "Could not open $treeFile";
    while( <TREE> ) {
        chomp;
        next if /^\s*$/; #ignore blank lines
        if( /^(\(.*\)\d+;)\s*$/ ) {
            $treeString = $1;
            last;
        } elsif( /^(.*);\s*$/ ) {
            die "Bad format for species file - missing root id or parentheses.";
        } else {
            die "Species tree string not found";
        }
    }
    close TREE;

    return teraUtils::parseNewick($treeString, 1); 
}
